;================================================================
;
;  DCF77 serial interface
;  version 1.0
;  
;  Microchip MPLAB IDE compiler   /www.microchip.com/
;
;================================================================

#include <P12F675.inc>

BG0_OFF EQU 0x2FFF
BG1_OFF EQU 0x1FFF

; __CONFIG BG0_OFF & BG1_OFF &  _CPD_OFF & _CP_OFF & _BODEN_OFF & _MCLRE_OFF & _PWRTE_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT
 __CONFIG _CPD_OFF & _CP_OFF & _BODEN_OFF & _MCLRE_OFF & _PWRTE_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT

#define IOPort GPIO       ; GPIO/PORTA
#define IOTris TRISIO     ; TRISIO, TRISA

#define dcfbuflen 0x0A
#define rxfbuflen 0x05

 cblock 0x20
  rxbuf: rxfbuflen
  rxchar, rxbit, txchar, txbit, rxtmr
  flag, dcfbit, HiTMR, HiTmrC, TmrS, TmrB, TmrDbg
  dcfbad1, dcfcnt1, dcfbuf1: dcfbuflen
  dcfbad0, dcfcnt0, dcfbuf0: dcfbuflen
  TMPD,CTMP
  DTMP1,DTMP2,DTMP3
  dcfflag, minute, hour, day, month, year
  dtvalidc, tmpval1, tmpval2, tmpc, tmpp, tmpp1
 endc



rxlength EQU 0x2
hitimer1 EQU .217  ; ~10ms ; .70 ~47ms

fdcfdata  EQU 0x0
fdcfmins  EQU 0x1
fdcfsync  EQU 0x2

fminp     EQU 0x4
fhourp    EQU 0x5
fdatep    EQU 0x6
fdatac    EQU 0x7

TMPW     EQU 0x5D
TMPS     EQU 0x5E
TMPFSR   EQU 0x5F

output0  EQU 0x0          ; GP0
DCFLed   EQU 0x1          ; GP1
DCFPon   EQU 0x2          ; GP2/INT
RXLine   EQU 0x3          ; GP3
DCFData  EQU 0x4          ; GP4
TXLine   EQU 0x5          ; GP5


#define USE_19200

#ifdef USE_9600
; **************    9600/ 104us    **************
delbit1  EQU .30          ;  33=104, 32=101, 31=97,6
delbit2  EQU .41          ;  49=151, 48=147, 47=145
#endif

#ifdef USE_19200
; **************    19200/ 52us    **************
delbit1  EQU .13            
delbit2  EQU .20            
#endif

#ifdef USE_38400
; **************    38400/ 26us    **************
delbit1  EQU .4          
delbit2  EQU .7            
#endif

cmp macro var1,const1
     movfw var1
     sublw const1
 endm


;================================================================
     ORG 0x3FF            ; OSCCAL
;     retlw 0x38           ; device depended value
;================================================================

     ORG     0x0
     goto    Start

     ORG     0x4     
Interrupt            
     movwf TMPW
     swapf STATUS,W
     bcf   STATUS,RP0
     movwf TMPS
     movfw FSR
     movwf TMPFSR
     btfsc INTCON,GPIF 
     goto  DOPCInt
     btfsc INTCON,INTF 
     goto  DoINT0
     btfsc INTCON,T0IF 
     goto  DoTimer0
     btfsc PIR1,TMR1IF 
     goto  DoTimer1

IEEnd                  
     movfw TMPFSR
     movwf FSR
     swapf TMPS,W
     movwf STATUS
     swapf TMPW,F
     swapf TMPW,W
     retfie


DoTimer0
     bcf   INTCON,T0IF
     goto  IEEnd

DoTimer1
     incf  HiTMR,w
     btfss STATUS,Z
     movwf HiTMR

     decf  rxtmr,w
     btfss STATUS,Z
     movwf rxtmr
     btfss STATUS,Z
     goto  DoTimer1Ex
     movfw rxbuf       
     btfsc STATUS,Z		
	 goto  DoTimer1Ex  
     clrf  rxbuf
DoTimer1Ex
     movlw hitimer1    
     movwf TMR1H
     bcf   PIR1,TMR1IF 
     goto  IEEnd

DOPCInt                
     btfsc IOPort,RXLine
     goto  DCFInt       
     call  USARTReceive
     movfw IOPort
     bcf   INTCON,GPIF  
     goto  IEEnd
DCFInt
     call  DCFReceive
     movfw IOPort
     bcf   INTCON,GPIF  
     goto  IEEnd

DoINT0                    
     bcf   INTCON,INTF  
     goto  IEEnd



DelayBit1_0         
     movlw delbit1  
     movwf TMPD     
N104 decfsz TMPD,F  
     goto  N104     
     return         

DelayBit1_5         
     movlw delbit2  
     movwf TMPD     
N156 decfsz TMPD,F  
     goto  N156     
     return         

SendChar            
     movwf txchar
     movlw .8
     movwf txbit
     bcf INTCON,GIE 
     bcf IOPort,TXLine
     call DelayBit1_0
SendChar0
     rrf txchar,F
     btfsc STATUS,C
     goto SendChar1
     bcf IOPort,TXLine
     goto SendChar2
SendChar1
     bsf IOPort,TXLine
SendChar2
     call DelayBit1_0
     decfsz txbit,F
     goto SendChar0
     bsf IOPort,TXLine
     bsf INTCON,GIE     
     call DelayBit1_0   
     call DelayBit1_0   
     return



InitTimer1      
     bcf   T1CON,TMR1ON 
     clrf  TMR1H
     clrf  TMR1L
     bcf   STATUS,RP0   
     movlw b'00000100' 
     movwf T1CON
     bsf   STATUS,RP0  
     bsf   PIE1,TMR1IE 
     bcf   STATUS,RP0  
     bcf   PIR1,TMR1IF 
     bsf   T1CON,TMR1ON
     return




Start
     bsf   STATUS,RP0 
     call  0x3FF
     movwf OSCCAL
     bcf   STATUS,RP0 
     clrf  rxbuf
     clrf  rxbit
     clrf  IOPort     
     movlw b'00000111'
     movwf CMCON
     bsf   STATUS,RP0 
     clrf  IOC
     bsf   IOC,RXLine 
     bsf   IOC,DCFData
     bcf   OPTION_REG,NOT_GPPU
     bcf   WPU,GP4 
     movlw 0
     movwf ANSEL   
     movlw b'11011000'
     movwf IOTris           
     bcf   STATUS,RP0 
     bsf   IOPort,TXLine
     bsf   IOPort,RXLine  
     bcf   IOPort,DCFData 
     bsf   IOPort,DCFLed  
     bsf   IOPort,DCFPon  
     bcf   IOPort,output0
     movlw .255
     movwf HiTMR
     movwf HiTmrC
     call  InitTimer1
     call  Delay100ms     
     clrf  flag
     call  DCFInit0
     call  DCFInit1
     clrf  dtvalidc
     call  DecodeTime     
     movlw b'11001000'    
     movwf INTCON 
     clrf  TmrDbg         
Loop
     clrwdt
     call  CheckUSART
     movfw TmrDbg
     btfsc STATUS,Z
     goto  Loop 

     movfw TmrS
     btfss STATUS,Z
     goto  LoopSend
     goto  Loop
LoopSend
     call  SendChar
     clrf  TmrS
     movfw IOPort
     andlw b'00010000'
     call  SendChar
     movfw TmrB     
     btfss STATUS,Z
     call  SendChar
     clrf  TmrB
     movfw dcfcnt0
     call  SendChar
     
     btfsc flag,fdcfdata
     call  SendDCFData
     goto  Loop


SendDCFData
     bcf   flag,fdcfdata
     call  GetTime1
     return
 
DCFReceive
     movfw HiTMR
     clrf  HiTMR
     movwf HiTmrC   
     movwf TmrS
     movlw hitimer1 
     movwf TMR1H

     btfss IOPort,DCFData 
     goto  DCFReceive1
DCFReceive0               
     bcf   flag,fdcfsync
     bcf   IOPort,DCFLed
     btfss flag,fdcfmins
     goto  DCFWaitMinute  
     movlw .5             
     subwf HiTmrC,W       
     btfss STATUS,C
     goto  DCFBadData
     movlw .15            
     subwf HiTmrC,W
     btfss STATUS,C
     goto  DCFBit0
DCFBit1
     movlw dcfbuf0
     addwf dcfbuf0,W 
     movwf FSR            
     movfw INDF
     iorwf dcfbit,w
     movwf INDF                
DCFBit0
     incf dcfcnt0,f
     bcf  STATUS,C
     rlf  dcfbit,f
     btfsc STATUS,C
     goto  DCFBitX
     return
DCFBitX
     movlw .1
     movwf dcfbit
     incf  dcfbuf0,f
     movlw dcfbuflen
     subwf dcfbuf0,w
     btfsc STATUS,C
     decf  dcfbuf0,f
     return
DCFReceive1              
     bsf   IOPort,DCFLed
     movlw .50           
     subwf HiTmrC,W      
     btfss STATUS,C     
     goto  DCFBadData    
     movlw .250          
     subwf HiTmrC,W      
     btfsc STATUS,C     
     goto  DCFBadData    
     movlw .150          
     subwf HiTmrC,W     
     btfsc STATUS,C     
     goto  DCFNewMinute  
     return
DCFWaitMinute
     movlw .3
     movwf TmrB
     return
DCFBadData               
     bcf   flag,fdcfmins
     incf  dcfbad0,w
     btfss STATUS,Z
     movwf dcfbad0
     movlw .1
     movwf TmrB
 
     return
DCFNewMinute    
     movlw .2
     movwf TmrB

     bsf  flag,fdcfmins
     movfw dcfbad0
     movwf dcfbad1
     movfw dcfcnt0
     movwf dcfcnt1
     movfw dcfbuf0
     movwf dcfbuf1
     movfw dcfbuf0+1
     movwf dcfbuf1+1
     movfw dcfbuf0+2
     movwf dcfbuf1+2
     movfw dcfbuf0+3
     movwf dcfbuf1+3
     movfw dcfbuf0+4
     movwf dcfbuf1+4
     movfw dcfbuf0+5
     movwf dcfbuf1+5
     movfw dcfbuf0+6
     movwf dcfbuf1+6
     movfw dcfbuf0+7
     movwf dcfbuf1+7
     movfw dcfbuf0+8
     movwf dcfbuf1+8
     movfw dcfbuf0+9
     movwf dcfbuf1+9
     call  DecodeTime     
     bsf   flag,fdcfdata

DCFInit0
     clrf  dcfbad0
     clrf  dcfcnt0
     movlw .1
     movwf dcfbit
     movlw dcfbuflen
     movwf CTMP
     movlw dcfbuf0		  
     movwf FSR            
DCFInit01
     clrf  INDF	         
     incf  FSR,F
     decfsz CTMP,F
     goto  DCFInit01
     incf  dcfbuf0,f
     return

DCFInit1
     clrf  dcfbad1
     clrf  dcfcnt1
     movlw dcfbuflen
     movwf CTMP
     movlw dcfbuf1		  
     movwf FSR            
DCFInit11
     clrf  INDF	         
     incf  FSR,F
     decfsz CTMP,F
     goto  DCFInit11
     return


USARTReceive
     movlw .50
     movwf rxtmr

     movlw .8             
     movwf rxbit          
     clrf  rxchar         
     call  DelayBit1_5
CheckUSARTReceive0
     bcf   STATUS,C       
     btfsc IOPort,RXLine  
     bsf   STATUS,C
     rrf   rxchar,F       
     call  DelayBit1_0    
     decfsz rxbit,F       
     goto  CheckUSARTReceive0
     movlw rxlength
     subwf rxbuf,W
     btfsc STATUS,C       
     return
     incf  rxbuf,F		  
     movlw rxbuf		  
     addwf rxbuf,W        
     movwf FSR            
     movfw rxchar         
     movwf INDF           
     return


CheckUSART
     movlw rxlength
     subwf rxbuf,W
     btfss STATUS,C       
     return
     bcf   INTCON,GPIE    
cmd0    
     cmp rxbuf+1,.0       
     btfss STATUS,Z		
	 goto  cmd1
     call  SendBuffer    
     goto  CheckUSARTEx
cmd1
     cmp rxbuf+1,.1
     btfss STATUS,Z		
	 goto  cmd2
     call  SendVersion
     goto  CheckUSARTEx
cmd2
     cmp rxbuf+1,.2
     btfss STATUS,Z		
	 goto  cmd3
     call  GetStatus
     goto  CheckUSARTEx
cmd3
     cmp rxbuf+1,.3
     btfss STATUS,Z		
	 goto  cmd4
     cmp rxbuf+2,.0
     btfss STATUS,Z		     
     call  WaitForSecSync
     call  GetTime0
     call  GetTime1
     call  SendLastTmr
     goto  CheckUSARTEx
cmd4
     cmp rxbuf+1,.4
     btfss STATUS,Z		
	 goto  cmd5
     cmp rxbuf+2,.0
     btfss STATUS,Z		
     call  WaitForSecSync
     call  SendDecodedTime
     goto  CheckUSARTEx
cmd5
     cmp rxbuf+1,.5
     btfss STATUS,Z		
	 goto  cmd6
     cmp rxbuf+2,.1
     btfss STATUS,Z		
	 bcf   IOPort,DCFPon   
     btfsc STATUS,Z		
	 bsf   IOPort,DCFPon   
     goto  CheckUSARTEx
cmd6
     cmp rxbuf+1,.6
     btfss STATUS,Z		
	 goto  cmd7
     movfw rxbuf+2
     movwf TmrDbg          
     call  SendChar
     goto  CheckUSARTEx
cmd7
     cmp rxbuf+1,.7
     btfss STATUS,Z		
	 goto  cmd8
     cmp rxbuf+2,.0
     btfss STATUS,Z		
     call  WaitForSecSync
     call  GetTime1
     call  SendLastTmr
     goto  CheckUSARTEx
cmd8
     cmp rxbuf+1,.8
     btfss STATUS,Z		
	 goto  cmd9

     goto  CheckUSARTEx
cmd9
     cmp rxbuf+1,.9
     btfss STATUS,Z		
	 goto  cmd10

     goto  CheckUSARTEx
cmd10
     
   
CheckUSARTEx
     call  SendCRLF
     clrf  rxbuf
     bsf   INTCON,GPIE  
     return


GetTime0
     movfw dcfbad0
     call  SendChar
     movfw dcfcnt0
     call  SendChar
     call  SendDCF0
     return

GetTime1
     movfw dcfbad1
     call  SendChar
     movfw dcfcnt1
     call  SendChar
     call  SendDCF1
     return

DecodeTime
     clrf  dcfflag
     clrf  minute
     clrf  hour
     clrf  day                
     clrf  month            
     clrf  year
     movfw dcfbad1
     btfss STATUS,Z
     goto  DecodeTimeEx     
     cmp   dcfcnt1,0x3B     
     btfss STATUS,Z
     goto  DecodeTimeEx     
     btfss dcfbuf1+.3,4
     goto  DecodeTimeEx     
     movfw dcfbuf1+.3
     andlw b'00001111'      
     movwf dcfflag

     swapf dcfbuf1+.3,W
     andlw b'00001110'
     movwf tmpval1
     bcf   STATUS,C
     rrf   tmpval1,F
     swapf dcfbuf1+.4,W
     andlw b'11110000'
     movwf tmpval2
     bcf   STATUS,C
     rrf   tmpval2,W
     iorwf tmpval1,F

     clrf  tmpp
     movlw .7
     call  DecodeValue   
     movfw tmpval2
     movwf minute

     movlw .1
     movwf tmpp1
     andwf tmpp,F
     movlw b'00010000'
     andwf dcfbuf1+.4,W
     btfsc STATUS,Z
     clrf  tmpp1
     movfw tmpp1
     bcf   flag,fminp
     bcf  STATUS,C
     subwf tmpp,W
     btfss STATUS,Z
     bsf   flag,fminp     

     swapf dcfbuf1+.4,W   
     andlw b'00001110'
     movwf tmpval1
     bcf   STATUS,C
     rrf   tmpval1,F
     swapf dcfbuf1+.5,W   
     andlw b'01110000'
     movwf tmpval2
     bcf   STATUS,C
     rrf   tmpval2,W
     iorwf tmpval1,F

     clrf  tmpp
     movlw .6 
     call  DecodeValue
     movfw tmpval2
     movwf hour

     movlw .1
     movwf tmpp1
     andwf tmpp,F
     movlw b'00001000'
     andwf dcfbuf1+.5,W
     btfsc STATUS,Z
     clrf  tmpp1
     movfw tmpp1
     bcf   flag,fhourp
     bcf  STATUS,C
     subwf tmpp,W
     btfss STATUS,Z
     bsf   flag,fhourp    

     clrf  tmpp
     swapf dcfbuf1+.5,W   
     andlw b'00001111'
     movwf tmpval1
     swapf dcfbuf1+.6,W   
     andlw b'00110000'
     iorwf tmpval1,F
     movlw .6
     call  DecodeValue
     movfw tmpval2
     movwf day
     movfw dcfbuf1+.6     
     andlw b'00011100'
     movwf tmpval1
     bcf   STATUS,C
     rrf   tmpval1,F
     bcf   STATUS,C
     rrf   tmpval1,F
     movlw .3
     call  DecodeValue
     swapf tmpval2,W
     movwf month
     btfsc dcfbuf1+.2,7   
     bsf   month,7        
     swapf dcfbuf1+.6,W   
     andlw b'00001110'
     movwf tmpval1
     bcf   STATUS,C
     rrf   tmpval1,F
     movfw dcfbuf1+.7
     andlw b'00000011'
     movwf tmpval2
     bcf   STATUS,C
     rlf   tmpval2,F
     bcf   STATUS,C
     rlf   tmpval2,F
     bcf   STATUS,C
     rlf   tmpval2,W
     iorwf tmpval1,F

     movlw .5
     call  DecodeValue
     movfw tmpval2
     iorwf month,F
     movfw dcfbuf1+.7
     movwf tmpval1
     bcf   STATUS,C
     rrf   tmpval1,F
     bcf   STATUS,C
     rrf   tmpval1,F
     swapf dcfbuf1+.8,W
     andlw b'00110000'
     movwf tmpval2
     bcf   STATUS,C
     rlf   tmpval2,F
     bcf   STATUS,C
     rlf   tmpval2,W
     iorwf tmpval1,F

     movlw .8
     call  DecodeValue
     movfw tmpval2
     movwf year
     movlw .1
     movwf tmpp1
     andwf tmpp,F
     movlw b'00000100'
     andwf dcfbuf1+.8,W
     btfsc STATUS,Z
     clrf  tmpp1
     movfw tmpp1
     bcf   flag,fdatep
     bcf  STATUS,C
     subwf tmpp,W
     btfss STATUS,Z
     bsf   flag,fdatep    

     bsf   flag,fdatac    
     movfw flag
     andlw b'11110000'
     iorwf dcfflag,F

     incf  dtvalidc,W
     btfss STATUS,Z
     movwf dtvalidc       
     btfsc flag,fminp
     clrf  dtvalidc
     btfsc flag,fhourp
     clrf  dtvalidc
     btfsc flag,fdatep
     clrf  dtvalidc
     return

DecodeTimeEx
     clrf  dtvalidc
     return


DecodeValue
     movwf tmpc
     clrf  tmpval2
DecodeValue0
     movlw HIGH GetBitVal
     movwf PCLATH
     movfw tmpc
     call  GetBitVal
     andwf tmpval1,W
     btfsc STATUS,Z
     goto  DecodeValue1
     movlw HIGH GetBCDVal
     movwf PCLATH
     movfw tmpc
     call  GetBCDVal
     addwf tmpval2,F
     incf  tmpp,F
DecodeValue1
     decfsz tmpc,F
     goto  DecodeValue0 
     return


WaitForSecSync
     bsf   flag,fdcfsync   
WaitForSecSyncCont
     btfss flag,fdcfsync
     return
;     clrwdt
     movlw .255
     subwf HiTMR,W
     btfss STATUS,Z        
     goto  WaitForSecSyncCont
     return


SendDecodedTime     
     movfw dcfflag
     call  SendChar
     movfw minute
     call  SendChar
     movfw hour
     call  SendChar
     movfw day
     call  SendChar
     movfw month
     call  SendChar
     movfw year
     call  SendChar
     movfw dcfcnt0        
     call  SendChar
     movfw dcfbad0        
     call  SendChar
SendLastTmr
     movfw HiTmrC         
     btfss IOPort,DCFData
     addwf HiTMR,W
     call  SendChar
     return


GetStatus
     clrw
     movfw IOPort
     call  SendChar
     return


SendBuffer
     movfw rxbuf
     movwf CTMP
     call SendChar
     movlw rxbuf		
     movwf FSR          
SendBuffer1
     incf  FSR,F
     movfw INDF	        
     call  SendChar     
     decfsz CTMP,F
     goto  SendBuffer1
     return

SendDCF0
     movlw dcfbuflen
     movwf CTMP
     movlw dcfbuf0	
     movwf FSR      
SendDCF01
     movfw INDF	    
     call  SendChar 
     incf  FSR,F
     decfsz CTMP,F
     goto  SendDCF01
     return

SendDCF1
     movlw dcfbuflen
     movwf CTMP
     movlw dcfbuf1		
     movwf FSR          
SendDCF11
     movfw INDF	        
     call  SendChar     
     incf  FSR,F
     decfsz CTMP,F
     goto  SendDCF11
     return


SendVersion
     clrf  CTMP
SendVersion1
     movlw HIGH version
     movwf PCLATH
     movfw CTMP
     call  version
     andlw 0xFF
     btfsc STATUS,Z
     return
     call  SendChar    
     incf  CTMP,F
     goto  SendVersion1

SendCRLF
     movlw .13
     call  SendChar
     movlw .10
     call  SendChar
     return


Delay_
     movwf DTMP1
N0   movlw .100         
     movwf DTMP2
N1   movlw .167         
     movwf DTMP3
N2   decfsz DTMP3,F                
     goto  N2
     decfsz DTMP2,F
     goto  N1
     decfsz DTMP1,F                
     goto  N0
     return

Delay100ms
     movlw .2
     call  Delay_
     return





     ORG 0x3C0
GetBitVal
     addwf PCL,F
     retlw .0
     retlw .1
     retlw .2
     retlw .4
     retlw .8
     retlw .16
     retlw .32
     retlw .64
     retlw .128


     ORG 0x3D0
GetBCDVal
     addwf PCL,F
     retlw .0
     retlw .1
     retlw .2
     retlw .4
     retlw .8
     retlw .10
     retlw .20
     retlw .40
     retlw .80


     ORG 0x3E0
version   
     addwf PCL,F
     dt "DFC77 v1.0",0

     


 END

