//*********************************************************
//
// DCF77 serial port service
//
// DOS, Windows, Linux/Unix
//
// version 1.0,  (c) Zelenka 2012
//
//*********************************************************

#include <time.h>
#include <stdio.h>
#include <signal.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <fcntl.h>
#include <errno.h>
#include <sys/timeb.h>

#ifdef __MSDOS__
 #include <io.h>
 #include <dos.h>
#else
#ifdef __WIN32__
 #include <windows.h>
 #include <dos.h>
#else
 #define __UNIX__
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
 #include <termios.h>
#ifndef __linux__
  #include <sys/pstat.h>
  #include <sys/modem.h>
#endif
#endif
#endif


#define OneStopBit 0
#define One5StopBits 1
#define TwoStopBits 2

#define Parity_None 0
#define Parity_Odd 1
#define Parity_Even 2

#define LCont_Stop 4
#define LCont_Parity_Enable 8
#define LCont_Parity_Select 16
#define LCont_Dlab 128

#define MCont_DTR 1
#define MCont_RTS 2
#define MCont_OUT1 4
#define MCont_OUT2 8

#define LStat_Data_Ready 1
#define LStat_Thre 32

#define bufsize 1000
#define CMDLENGTH 2
#define ASCII 1
#define MESASURE_TIME 60
#define ERATIME 1970
#define SecsPerDay     86400.0
#define SecsPerHour     3600.0
#define SecsPerMin        60.0
#define MsecPerDay  86400000.0
#define MsecPerHour  3600000.0
#define MsecPerMin     60000.0
#define MsecTime        1000.0

char weekday[][10] =
  {"Unknown","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"};

int MonthDays[2][12] =
     {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,
      31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

typedef char buffer[bufsize];
char exedir[255];
char exename[255];
char iniexename[255];
char logexename[255];
char datexename[255];
char comport[30];
char logstr[1024];
int baud=19200;
int century=2000;
int writelog=1;
int writedata=0;
int DCFSetTime=0;
int SynchronizeDiff=100;   // Synchronize time if difference is great than (ms)
int SynchronizeMaxDiff=12; // Synchronize time if difference is less than (hours)
int ValidDateTimeCount=3;
int ValidDateTime = 0;

struct TDCFTime {
  struct timeb rtime;
  int r,a1,z1,z2,a2,s,ho,mi,se,ms,d,m,y,dw,vt;
} TDCFTime, LastDateTime;

#ifdef __MSDOS__
 unsigned int Port_Adr, Reg_RX, Reg_TX, Reg_Int_En, Reg_Int_Id;
 unsigned int Reg_LCont, Reg_MCont, Reg_LStat, Reg_MStat, Reg_Fifo;
 int port;
#else
#ifdef __WIN32__
 HANDLE port;
#else
 int port;
#endif
#endif


void binb(char n, char *result)
{ char a, b, c[9];
 for (a=7; a >=0;a--) {
  b = 1 << a;
  c[7-a] = 48 + ((n & b) == b);
 }
 c[8] = 0;
 strcpy(result,c);
}

void rbin(char v, char *result)
{ char i,j,k,l;
 j = 0;
 for (i=0;i<8;i++) {
  k = 7-i;
  l = (1 << i);
  if (v & l) j = j + (1 << k);
 }
 binb(j,result);
}

int IsLeapYear(unsigned int y)
{ return(((y % 4) == 0) && (((y % 100) != 0) || ((y % 400) == 0)));}

// since 00:00:00 GMT, January 1, 1970
int GetDiff(struct TDCFTime d, struct TDCFTime *dd, double *t)
{double c;
 int i,l,r,h;
 l = IsLeapYear(d.y);
 c = 0;
 for (i=ERATIME;i<d.y;i++) {if (IsLeapYear(i)) c = c + 366; else c = c + 365;}
 for (i=1; i<d.m; i++) {c = c + MonthDays[l][i-1];}
 h=d.rtime.timezone/60;
 if (d.rtime.dstflag != 0 ) h = h - 1;
 c = ((c + d.d - 1) * SecsPerDay + d.se + (d.mi * 60.0) + ((d.ho+h) * 3600.0))*MsecTime + d.ms;
 c = (d.rtime.time*MsecTime+d.rtime.millitm) - c;
 *t = c;
 if (c < 0) { r = 0; c = c * (-1);} else r = 1;
 memset(dd,0,sizeof(TDCFTime));
 dd->d  = c / MsecPerDay;
 c = c - dd->d * MsecPerDay;
 dd->ho = c / (MsecPerHour); c = c - (dd->ho * MsecPerHour);
 dd->mi = c / (MsecPerMin); c = c - (dd->mi * MsecPerMin);
 dd->se = c / MsecTime;
 dd->ms = c - (dd->se * MsecTime);
 return r;
}

void DiffDateTime(struct TDCFTime *d, double diff)
{double c;
 struct tm *t;
 struct timeb dt;
 ftime(&dt);
 c = dt.time * MsecTime + dt.millitm - diff;
 d->rtime.time = c / MsecTime;
 d->rtime.millitm = c - (d->rtime.time * MsecTime);
 d->rtime.timezone = dt.timezone;
 d->rtime.dstflag = dt.dstflag;
 d->ms = d->rtime.millitm;
 t = localtime(&d->rtime.time);
 d->ho = t->tm_hour;
 d->mi = t->tm_min;
 d->se = t->tm_sec;
 d->y = 1900+t->tm_year;
 d->m = 1+t->tm_mon;
 d->d = t->tm_mday;
}


char *uppercase(char *b)
{ unsigned int i;
  for (i=0; i<strlen(b);i++) { b[i]=toupper(b[i]); } return b;
}

char *ltrim(char *b)
{ while ((strlen(b) > 0) && (isspace(b[0]))) {
   memcpy(b,b+1,strlen(b));
  } return b;
}

char *rtrim(char *b)
{ int i = strlen(b)-1;
  while (i >= 0) {
   if (isspace(b[i])) b[i] = 0; else break;
   i--;
  } return b;
}

char *trim(char *b)
{ rtrim(b); ltrim(b); return b; }

void datetimestr(char *s)
{time_t timer;
 struct tm *t;
 timer = time(NULL);
 t = localtime(&timer);
 sprintf(s,"%2.2d.%2.2d.%4.4d %2.2d:%2.2d:%2.2d ",t->tm_mday,1+t->tm_mon,1900+t->tm_year,t->tm_hour,t->tm_min,t->tm_sec);
}

void msdatetimestr(char *s)
{struct timeb t;
 int i,l,d,m,y,ho,mi,se,ms;
 long c,cd,cc;
 ftime(&t);
 if (t.dstflag != 0) t.timezone = t.timezone - 60;
 c = t.time - t.timezone*60;
 ms = t.millitm;
 cc = 0;
 y = ERATIME;
 l = IsLeapYear(y);
 cd = c / SecsPerDay;
 c = c - cd * SecsPerDay;
 while (cc + 365 + l <= cd) { cc = cc + 365 + l; y++; l = IsLeapYear(y); }
 l = IsLeapYear(y);
 m = 1;
 while (cc + MonthDays[l][m-1] <= cd) {
  cc = cc + MonthDays[l][m-1]; m++; if (m>12) { m--; break; }
 }
 d = cd - cc + 1;
 ho = c / (SecsPerHour); c = c - (ho * SecsPerHour);
 mi = c / (SecsPerMin);
 se = c - (mi * SecsPerMin);
 sprintf(s,"%.2d.%.2d.%.4d %.2d:%.2d:%.2d.%.3d ",d,m,y,ho,mi,se,ms);
}

void wlog(char *s)
{ FILE *f;
  char b[1024];

 datetimestr(b); strcat(b,s); strcat(b,"\n");
 printf("%s",b);
 if (!writelog) return;
 f = fopen(logexename,"a+");
 if (f) {
  fputs(b,f);
  fclose(f);
 } else {
  printf("Cannot open logfile, error %d\n",errno);
 }
}

void writetostringfile(char *s, char *filename)
{ FILE *f;
 f = fopen(filename,"a+");
 if (f) {
  fputs(s,f);
  fclose(f);
 } else {
 printf("Cannot open file %s, error %d\n",filename,errno);
 }
}


int str_readinifile(char *section, char *ident, char *result, char *def)
{ FILE *f;
 char s[255], p[255], SECTION[255],IDENT[255];
 int found;
 found = 0;
 strcpy(result,def);
 f = fopen(iniexename,"rb");
 if (f) {
   strcpy(SECTION,section); strcpy(IDENT,ident);
   uppercase(SECTION); uppercase(IDENT);
   while (!feof(f)) {
    if (fgets(s,sizeof(s)-1,f)) {
     trim(s); strcpy(p,s); uppercase(s);
     // remove \r\n...
     if ((s[strlen(s)-1] == 10) || (s[strlen(s)-1] == 13)) s[strlen(s)-1] = 0;
     if ((s[strlen(s)-1] == 10) || (s[strlen(s)-1] == 13)) s[strlen(s)-1] = 0;
     if ((!found) && (strstr(s,SECTION) == s)) { found = 1;}
     if (found) {
      if (strstr(s,IDENT) == s) {
       trim(strcpy(result,strstr(p,"=")+1));
       break;
      }
     }
    } else {
     if (!feof(f)) {
      sprintf(logstr,"Read error %d on %s",errno, iniexename);
      wlog(logstr);
     }
     return 0;
    }
   }
   fclose(f);
   return 1;
 } else {
  sprintf(logstr,"Can't open file %s",iniexename);
  wlog(logstr);
  return 0;
 }
}

int int_readinifile(char *section, char *ident, int *result, int def)
{ char s[255];
  int r;
  char *e;
 *result = def;
 r = str_readinifile(section, ident, s, "");
 if (!r) return 0;
 *result=strtol(s,&e,10);
 return 1;
}



int GetDosPort(void)
{ char cp[30];
 strcpy(cp,comport);
 uppercase(cp);
 if (strcmp(cp,"COM1") == 0) { return 0; } else
 if (strcmp(cp,"COM2") == 0) { return 1; } else
 if (strcmp(cp,"COM3") == 0) { return 2; } else
 if (strcmp(cp,"COM4") == 0) { return 3; } else { errno = 2; return -1; }
}

#ifdef __WIN32__
void WriteSysError(int err)
{ char txtbuf[256],shbuf[256];
  errno = err;
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, errno,
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), txtbuf, sizeof(txtbuf)-1,NULL);
  CharToOem(txtbuf,shbuf);
  sprintf(logstr,"%s", shbuf);
  wlog(logstr);
}
#endif

int setportcom(long Baud, int Data, int Parity, int Stop)
{
#ifdef __MSDOS__
unsigned char X, LSB, MSB;
unsigned int  Divisor;
 Port_Adr = peek(0x40, GetDosPort()*2);
 if (Port_Adr == 0) { return 0; }
 Reg_RX = Port_Adr;
 Reg_TX     = Port_Adr;
 Reg_Int_En = Port_Adr + 1;
 Reg_Int_Id = Port_Adr + 2;
 Reg_Fifo   = Port_Adr + 2;
 Reg_LCont  = Port_Adr + 3;
 Reg_MCont  = Port_Adr + 4;
 Reg_LStat  = Port_Adr + 5;
 Reg_MStat  = Port_Adr + 6;
 Divisor = (unsigned long)115200 / Baud;
 MSB = Divisor >> 8;
 LSB = Divisor & 0x0FF;
 outportb(Reg_LCont, LCont_Dlab);
 outportb(Port_Adr, LSB);
 outportb(Port_Adr+1, MSB);
 X = Data - 5;
 if (Stop == 2) X = X | LCont_Stop;
 if (Parity != 0)  X = X | LCont_Parity_Enable;
 if (Parity == Parity_Even) X = X | LCont_Parity_Select;
 outportb(Reg_LCont, X);
// outportb(Reg_MCont,MCont_OUT1+MCont_OUT2);
 outportb(Reg_Int_En,0);
 return 1;
#else
#ifdef __WIN32__
 DCB dcb;
 COMMTIMEOUTS t;
 memset(&dcb,0,sizeof(dcb));
 dcb.DCBlength = sizeof(dcb);
 GetCommState(port, &dcb);
 dcb.BaudRate = Baud;
 dcb.Parity = Parity;
 dcb.ByteSize = Data;
 dcb.StopBits = Stop;
 if (SetCommState(port, &dcb) ) {
  if (GetCommTimeouts(port,&t)) {
   t.ReadIntervalTimeout = 100;
   t.ReadTotalTimeoutMultiplier = 1;
   if (SetCommTimeouts(port, &t)) {
   } else { WriteSysError(GetLastError());  return 0; }
  } else { WriteSysError(GetLastError());  return 0; }
  return 1;
 } else { WriteSysError(GetLastError()); return 0; }
#else
 struct termios tio;
 unsigned int speed;
 if (tcgetattr(port,&tio) != -1) {

  tio.c_iflag = 0;
  tio.c_lflag = 0;
  tio.c_oflag = 0;

  tio.c_cflag = CREAD | CLOCAL;

  if (Stop == 2) tio.c_cflag |= CSTOPB;

  if (Parity) {
   tio.c_cflag |= PARENB;
   if (Parity == Parity_Even) tio.c_cflag |= PARODD;
  }// else tio.c_cflag &= ~PARENB;

  switch (Data) {
   case 5:  tio.c_cflag |= CS5; break;
   case 6:  tio.c_cflag |= CS6; break;
   case 7:  tio.c_cflag |= CS7; break;
   default: tio.c_cflag |= CS8;
  }

  switch (Baud) {
   case 1200:   speed=B1200; break;
   case 2400:   speed=B2400; break;
   case 4800:   speed=B4800; break;
   case 9600:   speed=B9600; break;
   case 19200:  speed=B19200; break;
   case 38400:  speed=B38400; break;
   case 57600:  speed=B57600; break;
   case 115200: speed=B115200; break;
   default: speed=B9600;
  }
  cfsetospeed(&tio, speed);
  cfsetispeed(&tio, speed);

  if (tcsetattr(port, TCSANOW, &tio)==0) { return 1; }

 } else { return 0; }
#endif
#endif
}

void setdtr(int state)
{ int ls;
#ifdef __MSDOS__
 ls = inportb(Reg_MCont);
 if (state != 0) { ls |= MCont_DTR; } else { ls &= ~MCont_DTR; }
 outportb(Reg_MCont,ls);
#else
#ifdef __WIN32__
 if (state != 0) { EscapeCommFunction(port, SETDTR); } else
 { EscapeCommFunction(port, CLRDTR); }
#else
 ioctl(port, TIOCMGET, &ls);
 if (state != 0) { ls |= TIOCM_DTR; } else { ls &= ~TIOCM_DTR; }
 ioctl(port, TIOCMSET, &ls);
#endif
#endif
}

void setrts(int state)
{ int ls;
#ifdef __MSDOS__
 ls = inportb(Reg_MCont);
 if (state != 0) { ls |= MCont_RTS; } else { ls &= ~MCont_RTS; }
 outportb(Reg_MCont,ls);
#else
#ifdef __WIN32__
 if (state != 0) { EscapeCommFunction(port, SETRTS); } else
 { EscapeCommFunction(port, CLRRTS); }
#else
 ioctl(port, TIOCMGET, &ls);
 if (state != 0) { ls |= TIOCM_RTS; } else { ls &= ~TIOCM_RTS; }
 ioctl(port, TIOCMSET, &ls);
#endif
#endif
}

int portread(buffer b, int l)
{ long j = 0;
  char bb[bufsize];
  char c;

 memset(bb,0,sizeof(bb));
#ifdef __MSDOS__
 while ((inportb(Reg_LStat) & LStat_Data_Ready) != 0) {
  c = inportb(Reg_RX);
  if (l >= sizeof(buffer)) { return 0; }
  b[l] = c; b[l+1] = 0;
  j++; l++;
 }
#else
#ifdef __WIN32__
 if (ReadFile(port, bb, sizeof(bb)-1,(LPDWORD) &j, NULL)) { memcpy(&b[l],bb,j); b[l+j]=0; }   //strcat(b,bb);
#else
 j = read(port, bb, sizeof(bb)-1); // -1 = errno otherwise positive number
 if (j <= 0) { j = 0; if (errno == EAGAIN) { errno = 0; } } else {  memcpy(&b[l],bb,j); b[l+j]=0; } //strcat(b,bb);
#endif
#endif
 return j;
}

int portwrite(buffer b)
{ long i,w;
 w = CMDLENGTH;
#ifdef __MSDOS__
 for (i=0; i < w; i++) {
  outportb(Reg_RX,b[i]);
  while ((inportb(Reg_LStat) & LStat_Thre) == 0) {}
 }
 i = w;
#else
#ifdef __WIN32__
 WriteFile(port, b, w, (LPDWORD) &i, NULL);
#else
 i = write(port, b, w);
#endif
#endif
 if (i == w) { return i; } else { return 0; }
}

int waitforresponse(buffer b, int timeout, int mc)
{ time_t t;
  int i,j;
 memset(b,0,sizeof(b));
 j = 0;
 t = time(NULL)+timeout;   // timeout in seconds
 while ((time(NULL) < t) && (j < mc)) {
  i = portread(b,j);  j = j + i;
  if ( i > 0) { t = time(NULL)+timeout;}
 }
 return j;
}

void printres(buffer b, int len, int ptype)
{ int x;
  char prs[1024];
  char s[10];
  strcpy(prs,"");
  for (x = 0; x < len; x++) {
   char c = b[x];
   if (ptype == ASCII) {
    if (c == '\r') sprintf(s,"<CR>"); else
    if (c == '\n') sprintf(s,"<LF>"); else sprintf(s,"%c",c);
   } else {
    sprintf(s,"%.2X ",c & (int) 0xFF);
   }
   strcat(prs,s);
  }
  sprintf(logstr,"%s",prs);
  wlog(logstr);
}

int sendstring(buffer b, int timeout, int mc)
{int r;
 buffer bb;
 portread(bb,0); // clear fifo
 if (portwrite(b)) {
  r = waitforresponse(b,timeout,mc);
  return(r);
 } else { return 0; }
}

int OpenPort(void)
{
#ifdef __MSDOS__
 port = GetDosPort();
#else
#ifdef __WIN32__
 port = CreateFile(comport, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
 errno = GetLastError();
#else
 port = open(comport,O_RDWR | O_NOCTTY | O_NONBLOCK ); //| O_NOCTTY | O_NDELAY
#endif
#endif
 return (int)port;
}

int ClosePort(void)
{ int r;
#ifdef __MSDOS__
  r = 0;
#else
#ifdef __WIN32__
  if (CloseHandle(port)) { r = 0; } else { r = 1; }
#else
  r = close(port);
#endif
#endif
 return r;
}

char GetValue(unsigned long c)
{char i,r,l;
 unsigned long k;
 char s[5];
 r = 0;
 for (i=0;i<8;i++) {
  k = (1 << i);
  sprintf(s,"%X",k);
  l = atoi(s);
  if ((c & k) == k) { r = r + l; }
 }
 return r;
}

int GetParity(unsigned long c, char l, char p)
{int i,n;
 unsigned long k;
 n = 0;
 for (i=0;i<l;i++) {
  k = (1 << i);
  if ((c & k) == k) { n++; }
 }
 return (p == (n % 2));
}

void UpdateSystemTime(struct TDCFTime d)
{ double t,ta;
  struct TDCFTime dd;
  char sign[1];
  int r;
#ifdef __UNIX__
//  struct timespec ts;
  struct timeval tv;
  struct timezone tz;
#else
  struct time ttime;
  struct date ddate;
  struct _SYSTEMTIME st;
#endif
 // if received time is "Sync" width DCF device:
 // add min 6ms for "Get Decoded Time"
 // add min 15ms for "Get Time Data"
 // add processor time ???ms or optimize program code :-)

  sprintf(logstr,"R:%d A1:%d Z1:%d Z2:%d A2:%d S:%d",d.r,d.a1,d.z1,d.z2,d.a2,d.s); wlog(logstr);
  sprintf(logstr,"%.2d:%.2d:%.2d.%.3d %.2d.%.2d.%.4d %s",d.ho,d.mi,d.se,d.ms,d.d,d.m,d.y,weekday[d.dw]); wlog(logstr);

  if (!d.vt) { ValidDateTime = 0; return; }
  if (d.ms >= 1000) d.ms = 0; // correct bad data
  if (d.dw == 7) d.dw = 0;
  if ((LastDateTime.y != d.y) ||
     (LastDateTime.m != d.m) ||
     (LastDateTime.dw != d.dw) ||
     (LastDateTime.d != d.d) ||
     (LastDateTime.ho != d.ho) ||
     (LastDateTime.mi != d.mi)) {  // only next minute
//    (LastDateTime.se != d.se) ||
//    (LastDateTime.ms != d.ms)) {
   LastDateTime.y = d.y;
   LastDateTime.m = d.m;
   LastDateTime.dw = d.dw;
   LastDateTime.d = d.d;
   LastDateTime.ho = d.ho;
   LastDateTime.mi = d.mi;
   LastDateTime.se = d.se;
   LastDateTime.ms = d.ms;
   ValidDateTime++;
  }

 if (GetDiff(d,&dd,&t)) strcpy(sign,"-"); else strcpy(sign,"");
 sprintf(logstr,"Time diff: %s%dd %.2d:%.2d:%.2d.%.3d ContValidTime: %d",
  sign,dd.d,dd.ho,dd.mi,dd.se,dd.ms,ValidDateTime);
 wlog(logstr);
 if (!DCFSetTime) return;
 if (ValidDateTime >= ValidDateTimeCount) {
  ta = t; if (ta < 0) ta=ta*(-1);
  if (SynchronizeDiff <= ta) {
   if (SynchronizeMaxDiff < (ta/MsecPerHour)) {
    sprintf(logstr,"Big difference, can't synchronize time..."); wlog(logstr);
   } else {
    DiffDateTime(&dd,t);
    sprintf(logstr,"Synchronizing time to %.2d.%.2d.%.4d %.2d:%.2d:%.2d.%.3d",dd.d,dd.m,dd.y,dd.ho,dd.mi,dd.se,dd.ms);
    wlog(logstr);
#ifdef __UNIX__
    if (gettimeofday(&tv,&tz) == 0) {
     tv.tv_sec=dd.rtime.time;
     tv.tv_usec=dd.rtime.millitm*1000;    // microseconds
//     ts.tv_sec=dd.rtime.time;
//     ts.tv_nsec=dd.rtime.millitm*1000*1000; // nanoseconds
//     r = clock_settime(CLOCK_REALTIME, &ts);  // cc -lrt dcf77.c -o dcf77
     r = settimeofday(&tv,&tz);
     if (r != 0) {
      sprintf(logstr,"Can't set time! errno %d",errno); wlog(logstr);
     }
    } else {
     sprintf(logstr,"Can't gettimeofday! errno %d",errno); wlog(logstr);
    }
#else
#ifdef __MSDOS__
    ttime.ti_min = dd.mi;
    ttime.ti_hour = dd.ho;
    ttime.ti_sec = dd.se;
    ttime.ti_hund = dd.ms/10;
    ddate.da_year = dd.y;
    ddate.da_mon = dd.m;
    ddate.da_day = dd.d;
    settime(&ttime);
    setdate(&ddate);
#else
    st.wYear = dd.y;
    st.wMonth = dd.m;
    st.wDayOfWeek =dd.dw;
    st.wDay = dd.d;
    st.wHour = dd.ho;
    st.wMinute = dd.mi;
    st.wSecond = dd.se;
    st.wMilliseconds = dd.ms;
    if (!SetLocalTime(&st))
     {sprintf(logstr,"Can't set time! errno %d",GetLastError()); wlog(logstr);};
#endif
#endif
//   LastUpdatedLocalTime = .....
   }
  }
 }
}

void showtime(buffer b)
{ unsigned int f,bb;
  struct TDCFTime d;
  ftime(&d.rtime); d.vt = 1; d.s=1;
  f = b[0];
  d.a1 = (f & 1) != 0;
  d.z1 = (f & 2) != 0;
  d.z2 = (f & 4) != 0;
  d.a2 = (f & 8) != 0;
  d.mi = b[1];
  d.ho = b[2];
  d.d = b[3];
  d.m = b[4] & 0x0F;
  d.dw = (b[4] & 0xF0) >> 4;
  d.r = (b[4] & 0x80) >> 7;
  d.y = century + b[5];
  d.se = b[6];
  bb = b[7];
  d.ms = b[8] * 10;
  if ((f & 128) == 0) {
   ValidDateTime = 0; sprintf(logstr,"Bad data!"); wlog(logstr);
  } else {
   if ((f & 16) == 16) {d.vt=0;sprintf(logstr,"Bad minute parity!");wlog(logstr);}
   if ((f & 32) == 32) {d.vt=0;sprintf(logstr,"Bad hour parity!");wlog(logstr);}
   if ((f & 64) == 64) {d.vt=0;sprintf(logstr,"Bad date parity!");wlog(logstr);}
   if (bb != 0) {d.vt=0;sprintf(logstr,"Bad marks in new minute!");wlog(logstr);}
   UpdateSystemTime(d);
  }
}

void decodetime(buffer b)
{ int p1,p2,p3,i;
  unsigned long dt;
  char t[12],p[255],s[255];
  struct TDCFTime d;

  ftime(&d.rtime); d.vt = 1;
  memcpy(&t[0],&b[12],sizeof(t));
  d.r = (t[4] & 128) != 0;
  d.a1 = (t[5] & 1) != 0;
  d.z1 = (t[5] & 2) != 0;
  d.z2 = (t[5] & 4) != 0;
  d.a2 = (t[5] & 8) != 0;
  d.s = (t[5] & 16) != 0;

  if ((b[12] == 0) && (b[13] == 0x3B) && writedata) {
   datetimestr(p);
   for (i=12;i<=23;i++) {sprintf(s,"%.2X ",(int)b[i] & 0xFF); strcat(p,s);}
   for (i=15;i<=23;i++) { rbin(b[i],s); strcat(p,s); strcat(p," ");}
   strcat(p,"\n"); writetostringfile(p,datexename);
  }
  if ((b[12] != 0) || (b[13] != 0x3B)) {
   ValidDateTime = 0; sprintf(logstr,"Bad data!");wlog(logstr);
  } else {
   if (d.s == 0) { d.vt=0;sprintf(logstr,"Bad start bit!");wlog(logstr); }
   if (b[0] != 0) { d.vt=0;sprintf(logstr,"Bad marks in new minute!");wlog(logstr); }
   d.se = b[1];
   memcpy(&dt,&t[5],sizeof(dt));
   dt = dt >> 5;
   d.mi = GetValue(dt & 127);
   p1 = (dt >> 7) & 1;
   if (!GetParity(dt & 127,7,p1)) { d.vt=0; sprintf(logstr,"Bad minute parity!");wlog(logstr); }
   dt = dt >> 8;
   d.ho = GetValue(dt & 63);
   p2 = (dt >> 6) & 1;
   if (!GetParity(dt & 63,6,p2)) { d.vt=0; sprintf(logstr,"Bad hour parity!");wlog(logstr); }

   dt = dt >> 7;
   d.d = GetValue(dt & 63);
   memcpy(&dt,&t[8],sizeof(dt));
   dt = dt >> 2;
   d.dw = (dt & 7);
   dt = dt >> 3;
   d.m = GetValue(dt & 31);
   dt = dt >> 5;
   d.y = century + GetValue(dt & 255);
   d.ms = b[24]*10;

   memcpy(&dt,&t[7],sizeof(dt));
   p3 = (dt >> 26) & 1;
   dt = (dt >> 4) & 0x3FFFFF;
   if (!GetParity(dt,22,p3)) { d.vt=0; sprintf(logstr,"Bad date parity!");wlog(logstr); }

   UpdateSystemTime(d);
  }
}

void CtrlCexit(int sig)
{
 printf("\nCtrl-C\n");
 setrts(0);
 setdtr(0);
 ClosePort();
 sleep(1);
 exit(3);
}

int c_break(void)
{
 CtrlCexit(3);
 return 3;
}

void dos_win_exename(char *s)
{char *r,*rn;
  strcpy(exedir,""); strcpy(exename,"");
  r=strrchr(s,'\\');
  if (r) {
   strncpy(exedir,s,r-s);
   rn=strrchr(s,'.');
   if (rn) {
    strncpy(exename,r+1,rn-r-1);
   } else {printf("Error argv[]!\n");}
  } else {printf("Error argv[]!\n");}
 sprintf(iniexename,"%s\\%s.ini",exedir,exename);
 sprintf(logexename,"%s\\%s.log",exedir,exename);
 sprintf(datexename,"%s\\%s.dat",exedir,exename);
}

#ifdef __UNIX__
void unix_exename(char *s)
{char d[1024];
 char *r;
 int i;
 strcpy(exedir,""); strcpy(exename,s);
#ifdef __linux__
 i=readlink("/proc/self/exe",d,sizeof(d)-1);
 if (i < 0) {sprintf(logstr,"ReadLink error, errno %d", errno);wlog(logstr); }
#else
 struct pst_status st;
  i = -1;
  if (pstat_getproc(&st,sizeof(st),0,getpid()) == -1) {
   sprintf(logstr,"Can't pstat_getproc, errno %d",errno);wlog(logstr);
  } else {
   if (pstat_getpathname(d,sizeof(d),&st.pst_fid_text) == -1) {
    sprintf(logstr,"Can't pstat_getpathname, errno %d",errno);wlog(logstr);
   } else {i = 0;}
  }
#endif
 if (i>=0) {
  r=strrchr(d,'/');
  if (r) { strncpy(exedir,d,r-d+1); strcpy(exename,r+1); } else { strcpy(exename,d); }
 }
 sprintf(iniexename,"%s%s.ini",exedir,exename);
 sprintf(logexename,"%s%s.log",exedir,exename);
 sprintf(datexename,"%s%s.dat",exedir,exename);
// printf("argv0: %s\nlink: %s\nexedir: %s\nexename: %s\nini: %s\nlog: %s\ndat: %s\n",s,d,exedir,exename,iniexename,logexename,datexename);
}
#endif

void getexename(char *s)
{
#ifdef __UNIX__
 unix_exename(s);
#else
 dos_win_exename(s);
#endif
}


int main(int argc, char *argv[])
{ buffer b;
  char c,tmp[255],tz[255];
  int r,l,ptype,w;
  int timer,tmrc,timer_enable;
  int TimerGetDecodedTime;
  char initsec[]="[init]";

#ifdef __MSDOS__
 ctrlbrk(c_break); setcbrk(1);
#else
 signal(SIGINT,CtrlCexit);
#endif
 printf("Ctrl-C to exit\n");

 getexename(argv[0]);
// dos COMx, windows \\.\COMx, linux /dev/ttySx  /dev/ttyUSBx, unix eg. /dev/tty0p0
 if (! str_readinifile(initsec,"port",comport,"")) {}
 if (! str_readinifile(initsec,"tz",tz,"")) {}
 if (! int_readinifile(initsec,"baud",&baud,baud)) {}
// timer is in minutes
 if (! int_readinifile(initsec,"timer",&timer,1)) {}
 if (! int_readinifile(initsec,"TimerEnable",&timer_enable,0)) {}
 if (! int_readinifile(initsec,"century",&century,century)) {}
 if (! int_readinifile(initsec,"WriteLog",&writelog,writelog)) {}
 if (! int_readinifile(initsec,"WriteData",&writedata,writedata)) {}
 if (! int_readinifile(initsec,"DCFSetTime",&DCFSetTime,DCFSetTime)) {}
 if (! int_readinifile(initsec,"SynchronizeDiff",&SynchronizeDiff,SynchronizeDiff)) {}
 if (! int_readinifile(initsec,"SynchronizeMaxDiff",&SynchronizeMaxDiff,SynchronizeMaxDiff)) {}
 if (! int_readinifile(initsec,"ValidDateTimeCount",&ValidDateTimeCount,ValidDateTimeCount)) {}
// if TimerGetDecodedTime = 0 then use GetTimeData
 if (! int_readinifile(initsec,"TimerGetDecodedTime",&TimerGetDecodedTime,0)) {}


 if (OpenPort() != -1) {
   errno = 0;
   sprintf(logstr,"%s opened",comport);
   wlog(logstr);

    if (!setportcom(baud,8,Parity_None,OneStopBit)) {
     errno = 253;
     sprintf(logstr,"Can't set %s",comport);
     wlog(logstr);
    } else {

      setrts(1);
      setdtr(1);
      sleep(1);  // time for capacitor charge

/*
	b[0] = 5; // DCF module On/Off
	b[1] = 1; // start DCF module
	b[1] = 0; // stop DCF module
	l = 2;
*/
      ptype = ASCII;
      b[0] = 1;  // print hw version info
      b[1] = 0;
      b[2] = 0;
      l = 12;
      r = sendstring(b, 2, l);
      if (r >= l) {
       printres(b,r,ptype);
        while (1==1) {
	  ptype = 0;
          if (TimerGetDecodedTime) {
           b[0] = 4; // get decoded time
           b[1] = 1; // 0 = nosync
           l = 10;
          } else {
           b[0] = 3; // get time data
           b[1] = 1; // 0 = nosync
           l = 26;
          }
	  b[1] = 0;
	  r = sendstring(b, 2, l);
	  if (r >= l) {
	    printres(b,r,ptype);
            if (TimerGetDecodedTime) showtime(b); else decodetime(b);
	  } else {
           sprintf(logstr,"Receive error!");
           wlog(logstr);
          }

         // wait here
         if (timer_enable) {
          tmrc = (MESASURE_TIME*timer);
          for (w=1;w<tmrc;w++) {
           sleep(1);
	   printf("%4d secs left \r",tmrc-w);
          }
         } else break;

	}
      } else {
       errno = 250;
       sprintf(logstr,"Receive error!");
       wlog(logstr);
      }
    }

   setrts(0);
   setdtr(0);
   if (ClosePort() != 0) {
     sprintf(logstr,"Can't close %s",comport);
     wlog(logstr);
   } else {
    sprintf(logstr,"%s closed",comport);
    wlog(logstr);
   }

 }  else {
  errno = 254;
  sprintf(logstr,"%s not opened",comport);
  wlog(logstr);
 }

 if (errno != 0) {
  sprintf(logstr,"Return errno %d", errno);
  wlog(logstr);
 }
 sleep(5);
 return errno;
}


