unit Unit2;

interface

uses windows, sysutils, classes, forms, winsock, unitx;

// Simple NTP client/server, auth isn't supported
// supported only ERA 0 (due to 2036)

type

  TNTPPacket = record
   mode_vn_li: byte;
   stratum: byte;
   poll: byte;
   precision: byte;
   root_delay: dword;
   root_dispersion: dword;
   reference_identifier: array [0..3] of char; //dword;
   reference_timestamp_secs: dword;
   reference_timestamp_frac: dword;
   originate_timestamp_secs: dword;
   originate_timestamp_frac: dword;
   receive_timestamp_secs: dword;
   receive_timestamp_frac: dword;
   transmit_timestamp_secs: dword;
   transmit_timestamp_frac: dword;
  end;

  TNTPSrvThread = class(TThread)
  public
   constructor Create;
   procedure Execute; override;
  end;

  TNTPCliThread = class(TThread)
  public
   ip: string;
   constructor Create(aip: string);
   procedure Execute; override;
  end;

var NTPSrvOnOff: boolean = false;
    NTPSrvIsRunning: boolean = false;
    NTPSrv: TNTPSrvThread;
    NTPCli: TNTPCliThread;
    WSA_Data: WSAData;
    NTP_Server: string = '127.0.0.1'; // loopback address

const

      NTP_SRV_RCV_TIMEOUT = 2000; // 2000ms
      NTP_CLI_RCV_TIMEOUT = 6000; // 6000ms
      NTP_CLI_RCV_RETRY = 1;
      NTP_PORT = 123;

      NTP_LI: byte = 0; // no warning
      NTP_VN = 3;       // version number
      NTP_MODE_SRV = 4; // server
      NTP_MODE_CLI = 3; // client
      NTP_SEC_PER_DAY = dword(86400);
      NTP_NOW_CORR = -2;

      NTP_FRAC_PART = 4294967296; // = 1 seconds in fraction part

procedure StartNTPServer;
procedure StopNTPServer;
procedure NTPClientQuery(ip: string);


implementation

uses unit1;


procedure AddStrLog(s: string);
begin
 if Form1 <> nil then Form1.AddStr(s);
end;


function UTC_OFFSET_SEC: integer;
var r: dword;
    tz: TIME_ZONE_INFORMATION;
begin
 r := GetTimeZoneInformation(tz);
 if r = TIME_ZONE_ID_UNKNOWN then result := -3600 // CET
  else result := tz.Bias * 60;
end;

procedure TimeToNTP(n: TDateTime; var sec, fra: dword);
var v: Double;
begin
 v := (n + NTP_NOW_CORR) * NTP_SEC_PER_DAY + UTC_OFFSET_SEC;
 sec := htonl(dword(Trunc(v)));
 fra := htonl(dword(Trunc(Frac(v) * NTP_FRAC_PART)));
end;

function NTPToTime(sec, fra: dword): TDateTime;
begin
 result := (dword(ntohl(sec)) - dword(UTC_OFFSET_SEC) +
  (dword(ntohl(fra)) / NTP_FRAC_PART))/ NTP_SEC_PER_DAY - NTP_NOW_CORR;
end;


function ValidIP(ca: TSockAddrIn): boolean;
var s: string;
begin
 result := true;
 s := inet_ntoa(ca.sin_addr);
// if s = '127.0.0.1' then begin
//  AddStrLog('Srv QueryFrom '+s+' is not allowed!');
//  result := false;
// end;
end;

procedure NTPServer;
var s: TSocket;
    sa,ca: TSockAddrIn;
    r,l,n: integer;
    p: TNTPPacket;
    msg: array [0..1024] of char;
    t: TDateTime;
begin
 if NTPSrvIsRunning then exit;
 NTPSrvIsRunning := true;
try
 s := socket(AF_INET,SOCK_DGRAM,0);
 l := 1;
 r := setsockopt(s,SOL_SOCKET,SO_REUSEADDR,@l,sizeof(l));
 if r <> 0 then begin
  AddStrLog('Srv SetSockOpt reuseaddr: '+SysErrMsg);
 end;
 fillchar(sa,sizeof(sa),0);
 sa.sin_family := AF_INET;
 sa.sin_addr.s_addr := htonl(INADDR_ANY);
 sa.sin_port := htons(NTP_PORT);
 l := NTP_SRV_RCV_TIMEOUT;
 r := setsockopt(s,SOL_SOCKET,SO_RCVTIMEO,@l,4);
 if r <> 0 then begin
  AddStrLog('Srv SetSockOpt rcvtimeo: '+SysErrMsg);
 end;
 r := bind(s,sa,sizeof(sa));
 if r <> 0 then begin
  AddStrLog('Srv Bind: '+SysErrMsg);
 end;

 repeat
  if s <> INVALID_SOCKET then begin
   fillchar(msg,sizeof(msg),0);
   l := sizeof(ca);
   n := recvfrom(s,msg,sizeof(msg),0,ca,l);
   if (n > 0) then begin
    t := Now;
    AddStrLog('Srv QueryFrom '+inet_ntoa(ca.sin_addr)+':'+inttostr(ca.sin_port));
    if n <> sizeof(p) then AddStrLog('Packet length '+inttostr(n)+'!');

    if ValidIP(ca) then begin // exclude|include address implementation

     move(msg,p,sizeof(p));
     NTP_LI := 0;
     p.mode_vn_li := (NTP_LI shl 6) + (NTP_VN shl 3) + NTP_MODE_SRV;
     p.stratum := 1;       // primary reference (DCF77)
     p.poll := 6;          // 6=64s, 8=256s, 14=16284s (4h)
     p.precision := $EE;   //FA=0,015625s, F4=0,000244, EE=0,000004s
     p.root_delay := 0;
     p.root_dispersion := 0;
     p.reference_identifier := 'DCF'#0; // ie. inet_addr(PChar(local_ip));

     // time when local time was updated
     if LastUpdatedLocalTime = 0 then begin // not updated yet
      p.reference_timestamp_secs := 0;
      p.reference_timestamp_frac := 0;
     end else
     TimeToNTP(LastUpdatedLocalTime,p.reference_timestamp_secs,p.reference_timestamp_frac);
     // Originate timestamp is the local time at which the
     // request departed the client for the server
     p.originate_timestamp_secs := p.transmit_timestamp_secs;
     p.originate_timestamp_frac := p.transmit_timestamp_frac;
     // client's request arrived at the server
     TimeToNTP(t,p.receive_timestamp_secs,p.receive_timestamp_frac);
     t := Now;
     // departure time for the client
     TimeToNTP(t,p.transmit_timestamp_secs,p.transmit_timestamp_frac);

     l := sizeof(p);
     r := sendto(s,p,l,0,ca,sizeof(ca));
     if r <> l then begin
      AddStrLog('Srv SendTo: '+SysErrMsg);
     end;
    end;
   end else begin
    if n = SOCKET_ERROR then begin
     r := WSAGetLastError();
     if (r <> WSAEWOULDBLOCK) and (r <> WSAETIMEDOUT) then begin  // 10057 WSAENOTCONN
      AddStrLog('Srv RecvFrom: '+SysErrMsg);
     end;
    end;
   end;

  end;
 until not NTPSrvOnOff or application.Terminated;
 closesocket(s);
finally
 NTPSrvIsRunning := false;
end;
end;

procedure NTPGetTime(ip: string);
var s: TSocket;
    sa: TSockAddrIn;
    r,l,trc: integer;
    p: TNTPPacket;
    msg: array [0..1024] of char;
    ts,tr,tra,tre,rtt,off: TDateTime;
    systime: TSystemTime;
label tryrecv;
begin
 s := socket(AF_INET,SOCK_DGRAM,0);
 fillchar(sa,sizeof(sa),0);
 sa.sin_family := AF_INET;
 sa.sin_addr.s_addr := inet_addr(PChar(ip));
 sa.sin_port := htons(NTP_PORT);
 fillchar(p,sizeof(p),0);
 NTP_LI := 0;
 p.mode_vn_li := (NTP_LI shl 6) + (NTP_VN shl 3) + NTP_MODE_CLI;
 p.stratum := 5;
 p.poll := 6;
 p.precision := $EE;
 dword(p.reference_identifier) := sa.sin_addr.s_addr;

 ts := Now; // T1
 TimeToNTP(ts,p.transmit_timestamp_secs,p.transmit_timestamp_frac);
 l := sizeof(p);
 r := sendto(s,p,l,0,sa,sizeof(sa));
 if r = l then begin
   l := NTP_CLI_RCV_TIMEOUT;
   r := setsockopt(s,SOL_SOCKET,SO_RCVTIMEO,@l,4);
   if r <> 0 then begin
    AddStrLog('Cli SetSockOpt rcvtimeo: '+SysErrMsg);
   end;
   l := sizeof(sa);
   fillchar(msg,sizeof(msg),0);
   trc := 0;
tryrecv:
   r := recvfrom(s,msg,sizeof(msg),0,sa,l);
   if r > 0 then begin
    tr := Now; // T4
    move(msg,p,sizeof(p));
    tra := NTPToTime(p.transmit_timestamp_secs,p.transmit_timestamp_frac); // T3
    tre := NTPToTime(p.receive_timestamp_secs,p.receive_timestamp_frac); // T2
    rtt := (tr - ts) - (tre - tra);
    off :=  (tre+tra-ts-tr)/2;
    AddStrLog('Cli ReplyFrom '+inet_ntoa(sa.sin_addr)+':'+
     inttostr(sa.sin_port)+', time diff '+
     inttostr(trunc(off))+'d '+formatdatetime('hh:nn:ss.zzz',off)+
     ', RTT: '+formatdatetime('ss.zzz',rtt));

// Originate timestamp is the local time at which the request departed the client for the server
// Receive timestamp is the client's request arrived at the server
// Transmit timestamp is the departure time for the client
//-------------------------------------------------------------------------
// Originate Timestamp     T1   time request sent by client
// Receive Timestamp       T2   time request received at server
// Transmit Timestamp      T3   time reply sent by server
// Destination Timestamp   T4   time reply received at client

// The roundtrip delay d and local clock offset t are defined as
// d = (T4 - T1) - (T2 - T3)
// t = ((T2 - T1) + (T3 - T4)) / 2

    ts := Now + off;
    DateTimeToSystemTime(ts, systime);
    if CliSetNTPTime then begin
     AddStrLog('Cli Setting time to '+formatdatetime('dd.mm.yyyy hh:nn:ss.zzz',ts));
     if not SetLocalTime(systime) then begin
      AddStrLog('Cli SetLocalTime: '+SysErrMsg);
     end;
    end;

   end else begin
    AddStrLog('Cli RecvFrom: '+SysErrMsg);
    inc(trc);
    if trc < NTP_CLI_RCV_RETRY then begin
     AddStrLog('Cli is still waiting... ');
     goto tryrecv;
    end;
   end;
 end else AddStrLog('Cli SendTo: '+SysErrMsg);
 closesocket(s);
end;


constructor TNTPSrvThread.Create;
begin
 inherited Create(true);
 NTPSrvOnOff := true;
 freeonterminate := true;
 priority := tpNormal;
 resume;
end;

procedure TNTPSrvThread.Execute;
begin
try
 NTPServer;
except
 on e: exception do AddStrLog('UDPSrv '+e.message);
end;
end;


constructor TNTPCliThread.Create(aip: string);
begin
 inherited Create(true);
 ip := aip;
 freeonterminate := true;
 priority := tpNormal;
 resume;
end;

procedure TNTPCliThread.Execute;
begin
try
 NTPGetTime(ip);
except
 on e: exception do AddStrLog('UDPCli '+e.message);
end;
end;


procedure StartNTPServer;
begin
 if not NTPSrvIsRunning then begin
  NTPSrv := TNTPSrvThread.Create;
  AddStrLog('NTP server started');
 end else AddStrLog('NTP server is already started');
end;

procedure StopNTPServer;
var i: integer;
const slp=200;
begin
 if NTPSrvIsRunning then begin
  AddStrLog('Stopping NTP server...');
  NTPSrvOnOff := false;
  i := 0;
  while NTPSrvIsRunning do begin
   SleepEx(slp,true); inc(i);
   application.ProcessMessages;
   if i > 2*(NTP_SRV_RCV_TIMEOUT div slp) then break;
  end;
  if not NTPSrvIsRunning then AddStrLog('NTP server stopped')
   else AddStrLog('NTP server is still running');
 end else AddStrLog('NTP server is already stopped');
end;

procedure NTPClientQuery(ip: string);
begin
 NTPCli := TNTPCliThread.Create(ip);
end;

initialization
 WSAStartUp($101,WSA_Data);
finalization
 WSACleanup;

end.

